#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3			// Use modern global access method and strict wave access.
#pragma IgorVersion = 6.37	// Not tested in earlier versions

// This version is for Igor Pro 6.37 through 8.0x. A different version will ship with Igor Pro 9.00 and later.

#pragma IndependentModule = ScaleMarkerSizeInAxisUnits

// See the "Scale Marker Size in Axis Units.pxp" example experiment for background information.

static Function GetPlotAreaSpanInPoints(graphName, axisName)
	String graphName
	String axisName
	
	String axInfoStr = AxisInfo(graphName, axisName)
	String axTypeStr = StringByKey("AXTYPE", axInfoStr)
	Variable isVerticalAxis = CmpStr(axTypeStr,"left")==0 || CmpStr(axTypeStr,"right")==0
	
	GetWindow $graphName, psize				// Reads graph plot area dimensions in points
	Variable plotAreaMinInPoints, plotAreaMaxInPoints
	if (isVerticalAxis)
		plotAreaMinInPoints = V_bottom
		plotAreaMaxInPoints = V_top
	else
		plotAreaMinInPoints = V_left
		plotAreaMaxInPoints = V_right
	endif
	
	Variable plotAreaSpanInPoints = plotAreaMaxInPoints - plotAreaMinInPoints
	if (plotAreaSpanInPoints < 0)
		// The axis reversed
		plotAreaSpanInPoints *= -1
	endif
	
	return plotAreaSpanInPoints
End

// GetZMinAndZMax(graphName, axisName, traceName, zMin, zMax)
// Returns the zMin and zMax values to use with ModifyGraph zmrkSize assuming that
// the mrkMin and mrkMax parameters are 1 and 200.
// axisName must be the name of a built-in axis: left, right, bottom, or top.
// It can not be a free axis because there is currently no way to determine
// the size in points of the plot area associated with a free axis.
static Function GetZMinAndZMax(graphName, axisName, traceName, zMin, zMax)
	String graphName
	String axisName
	String traceName
	Variable& zMin				// Output
	Variable& zMax				// Output
	
	DoUpdate/W=$graphName		// Needed for GetAxis to return correct values
	
	GetAxis/W=$graphName/Q $axisName		// Sets V_min, V_max
	Variable minAxisValue=V_min, maxAxisValue=V_max
	Variable axisRange = maxAxisValue - minAxisValue
	if (axisRange < 0)
		// The axis reversed
		axisRange *= -1
	endif
	
	// Determine the size in points of the plot area associated with the axis
	Variable plotAreaSpanInPoints = GetPlotAreaSpanInPoints(graphName, axisName)
	
	// Find the axis units corresponding to 200 points of the axis
	Variable radiusFor200Points = axisRange * (200 / plotAreaSpanInPoints)
	
	zMax = radiusFor200Points
	zMin = zMax / 200
	
	// For debugging only
	// Printf "axisRange=%g, plotAreaSpanInPoints=%g, radiusFor200Points=%g, zMin=%g, zMax=%g\r", axisRange, plotAreaSpanInPoints, radiusFor200Points, zMin, zMax
End

static StrConstant kScaleMarkerSizeUserDataName = "ScaleMarkerSizeInAxisUnitsInfo"

static Function SetUserData(graphName, traceName, axisName)
	String graphName, traceName, axisName
	
	// Save the plot area size which UpdateTraceZMarkerSize uses to determine
	// if we need to update traces when the graph is resized.
	Variable plotAreaSpanInPoints = GetPlotAreaSpanInPoints(graphName, axisName)
	
	// Save the axis span which UpdateTraceZMarkerSize uses to determine
	// if we need to update traces when the the axis range is changed.
	GetAxis/W=$graphName/Q $axisName
	Variable axisMin = V_min, axisMax = V_max, axisSpan = abs(axisMax - axisMin)
	
	Variable currentTicks = ticks
	
	String userData
	sprintf userData, "PSIZE:%g;AXIS:%s;AXISSPAN:%g;LASTUPDATETICKS:%d;", plotAreaSpanInPoints, axisName, axisSpan, currentTicks

	ModifyGraph/W=$graphName userData($traceName)={$kScaleMarkerSizeUserDataName, 0, userData}
End

Function StopScalingMarkerSize(graphName, traceName)						// Accessible to client code
	String graphName, traceName
	
	ModifyGraph/W=$graphName userData($traceName)={$kScaleMarkerSizeUserDataName, 0, ""}
End
 
Function ScaleMarkerSize(graphName, axisName, traceName, radiusWave)	// Accessible to client code
	String graphName
	String axisName
	String traceName
	Wave radiusWave

	// zmrkSize($traceName)={markerSizeWave,zMin,zMax,minMarkerSize,maxMarkerSize}
	Variable zMin, zMax
	GetZMinAndZMax(graphName, axisName, traceName, zMin, zMax)
	ModifyGraph/W=$graphName zmrkSize($traceName)={radiusWave,zMin,zMax,1,200}
	
	// This information is used by UpdateTraceZMarkerSize to update the trace if the axis range is changed.
	SetUserData(graphName, traceName, axisName)
	
	// Set window hook so we can update the zmrkSize settings when the axis range is changed
	SetWindow $graphName, hook(ScaleMarkerSizeWindowHook) = ScaleMarkerSizeWindowHook
	
	// Printf "zMin=%g, zMax=%g\r", zMin, zMax	// For debugging only
End

static Function/WAVE GetZMarkerSizeWave(graphName, traceName)
	String graphName, traceName
	
	String info = TraceInfo(graphName, traceName, 0)
	Variable infoLen = strlen(info)
	Variable pos = strsearch(info, "RECREATION:", 0)
	pos += strlen("RECREATION:")
	String recreationInfo = info[pos,infoLen]
	String zmrkSizeInfo = StringByKey("zmrkSize(x)", recreationInfo, "=")	// e.g., {::radiusData,0,1,1,200}
	if (CmpStr(zmrkSizeInfo[0],"{") != 0)
		// This happens if the user turns zmrkSize mode off
		return $""
	endif
	
	String relativePath = StringFromList(0, zmrkSizeInfo[1,10000], ",")
	// Print relativePath					// For debugging only
	Wave/Z radiusWave = $relativePath
	return radiusWave
End

static Function PlotSizeChanged(graphName, axisName, prevPlotSize)
	String graphName
	String axisName
	Variable prevPlotSize
	
	Variable currentPlotSize = GetPlotAreaSpanInPoints(graphName, axisName)
	Variable plotAreaDif = abs(currentPlotSize - prevPlotSize)
	if (plotAreaDif < .001*currentPlotSize)		// Filter out differences because of roundoff
		// The plot area size did not materially change since the last update
		return 0
	endif
	
	return 1
End

static Function AxisSpanChanged(graphName, axisName, prevAxisSpan)
	String graphName
	String axisName
	Variable prevAxisSpan
	
	GetAxis/W=$graphName/Q $axisName
	Variable axisMin = V_min, axisMax = V_max
	Variable currentAxisSpan = abs(axisMax - axisMin)
	if (currentAxisSpan==0 && prevAxisSpan==0)
		// The axis min and max are the same and were the same the last time we updated
		return 0
	endif
	
	Variable axisSpanDif = abs(currentAxisSpan - prevAxisSpan)
	if (axisSpanDif < .001*currentAxisSpan)		// Filter out differences because of roundoff
		// The axis did not materially change since the last update
		return 0
	endif
	
	return 1
End

static Function UpdateTraceZMarkerSize(graphName, traceName, userData, eventCode)
	String graphName, traceName, userData
	Variable eventCode			// 6=resize, 8=modified
	
	Wave/Z radiusWave = GetZMarkerSizeWave(graphName, traceName)
	if (!WaveExists(radiusWave))
		// This happens if the user turns zmrkSize mode off
		return -1	
	endif
	
	String axisName = StringByKey("AXIS", userData)
	
	Variable prevPlotSize = NumberByKey("PSIZE", userData)
	Variable pSizeChanged = PlotSizeChanged(graphName, axisName, prevPlotSize)
		
	Variable prevAxisSpan = NumberByKey("AXISSPAN", userData)
	Variable axisChanged = AxisSpanChanged(graphName, axisName, prevAxisSpan)
	
	// Don't modify the trace if neither the plot size nor the axis span changed
	// because that would create another modified event and call our hook again
	// and the process would repeat forever.
	if (!pSizeChanged && !axisChanged)
		// Printf "Update not needed for trace %s", traceName		// For debugging only
		return 0
	endif
	
	// We expect that a modify event is sent because of a resize event
	// because we call ModifyGraph from the resize event. That is OK,
	// because we are not triggering another resize event. What we need
	// to block is a modify event caused by a modify event since calling
	// ModifyGraph in that case would lead to an unending chain of modified events.
	if (eventCode==8 && axisChanged)		// Repeat modified event due to axis change?
		// This prevents cascading modified events if the logic above fails for some unknown reason
		Variable lastUpdateTicks = NumberByKey("LASTUPDATETICKS", userData)
		Variable currentTicks=ticks, ticksDif=currentTicks-lastUpdateTicks
		if (ticksDif <= 2)
			// This should not happen
			Printf "UpdateTraceZMarkerSize running too frequently, ticks dif=%d\r", ticksDif
	 		return 0
		endif
	endif
	
	ScaleMarkerSize(graphName, axisName, traceName, radiusWave)
	
	//	Printf "Updated Z marker size settings for trace %s\r", traceName	// For debugging only ***
	return 0
End

static Function UpdateZMarkerSize(graphName, eventCode)	// Updates any traces in graph that we are tracking
	String graphName
	Variable eventCode			// 6=resize, 8=modified
	
	Variable numTracesBeingScaled = 0
	
	String list = TraceNameList(graphName, ";", 1)
	Variable numTraces = ItemsInList(list)
	Variable i
	for(i=0; i<numTraces; i+=1)
		String traceName = StringFromList(i, list)
		String userData = GetUserData(graphName, traceName, kScaleMarkerSizeUserDataName)
		if (strlen(userData) > 0)
			Variable err = UpdateTraceZMarkerSize(graphName, traceName, userData, eventCode)
			if (err == 0)
				numTracesBeingScaled += 1
			else
				// This happens if the user turns zmrkSize mode off. We stop tracking this trace.
				StopScalingMarkerSize(graphName, traceName)
				// Printf "Cleared user data for trace %s\r", traceName	// For debugging only ***
			endif
		endif
	endfor
	
	return numTracesBeingScaled
End

Function ScaleMarkerSizeWindowHook(s)
	STRUCT WMWinHookStruct &s

	Variable hookResult = 0
	
	String graphName = s.winName
	Variable eventCode = s.eventCode
	switch(eventCode)
		case 6:				// Resize event
		case 8:				// Modified event
			// Printf "UpdateZMarkerSize is running, eventCode=%d\r", eventCode	// For debugging only ***
			Variable numTracesBeingScaled = UpdateZMarkerSize(graphName, eventCode)
			if (numTracesBeingScaled == 0)
				// The graph no longer has traces that we are scaling so disconnect our hook function
				SetWindow $graphName, hook(ScaleMarkerSizeWindowHook) = $""
				// Printf "Cleared ScaleMarkerSizeWindowHook hook function for graph %s\r", graphName	// For debugging only ***
			endif
			break
	endswitch

	return hookResult		// 0 if nothing done, else 1
End
